//----------------------------------------------------------------------------
//
// File:        ti-disk.hpp
// Date:        27-Mar-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1998-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#ifndef _TIDISK_HPP_
#define _TIDISK_HPP_

#define REG_STATUS		0x5FF0
#define REG_RD_TRACK		0x5FF2
#define REG_RD_DATA		0x5FF6
#define REG_COMMAND		0x5FF8
#define REG_WR_TRACK		0x5FFA
#define REG_WR_SECTOR		0x5FFC
#define REG_WR_DATA		0x5FFE

// Not Ready
// Seek Error ( track not found )
// Record Not Found ( track/sector/side not found )
// Lost Data
// CRC Error
// Write protect

#define STATUS_NOT_READY	0x80		// No Diskette/Drive or Bad H/S
#define STATUS_WRITE_PROTECTED	0x40		// Write Protected Disk
#define STATUS_UNKNOWN_3	0x20
#define STATUS_SEEK_ERROR	0x10		// Seek error or Bad H/S
#define STATUS_CRC_ERROR	0x08		// CRC Error on a read
#define STATUS_LOST_DATA	0x04		// Lost data on a read
#define STATUS_INDEX_PULSE	0x02
#define STATUS_UNKNOWN_8	0x01

#define STATUS_NOT_FOUND	( 0x08 | 0x04 )

typedef unsigned char BYTE;

struct sSectorInfo {
    BYTE   ActualCylinder;
    BYTE   ActualSide;
    BYTE   LogicalCylinder;
    BYTE   LogicalSide;
    BYTE   LogicalSector;
    BYTE   Length;
    USHORT DataCount;
};

struct sSector {
    sSectorInfo  Info;
    BYTE        *Data;
};

struct sTrack {
    int          noSectors;
    sSector      Sector [9];
};

struct sDiskInfo {
    int          noTracks;
    sTrack       Track [80];
};

class cDisk : public cDevice {

    enum TRAP_TYPE_E {
        TRAP_DISK
    };

    UCHAR         m_Status;
    UCHAR         m_LastData;

    USHORT        m_DriveSelect;
    USHORT        m_HeadSelect;
    USHORT        m_TrackSelect;
    USHORT        m_TrackIndex;
    USHORT        m_SectorIndex;
    bool          m_TransferEnabled;
    bool          m_WriteTrack;

    bool          m_Dirty [3];
    const char   *m_DiskFile [3];
    sDiskInfo     m_DiskImage [3];
    sDiskInfo    *m_CurDisk;
    sSector      *m_CurSector;

    int           m_DataBytes;
    BYTE          m_AddressBuffer [6];
    BYTE          m_TrackBuffer [0x1000];
    BYTE         *m_DataPtr;

    UCHAR         m_TrapIndex;

    void FindSector ();
    void HandleCommand ( USHORT );

    UCHAR ReadByte ();
    void  WriteByte ( UCHAR );

    void Restore ();
    void Seek ();
    void StepIn ();
    void StepOut ();
    void ReadSector ();
    void WriteSector ();
    void ReadAddress ();
    void ReadTrack ();
    void WriteTrack ();
    void WriteTrackComplete ();

    static USHORT TrapFunction ( void *, int, bool, const ADDRESS, USHORT );

    USHORT WriteMemory ( const ADDRESS, USHORT );
    USHORT ReadMemory ( const ADDRESS, USHORT );

public:

    cDisk ();
    ~cDisk ();

    void LoadDisk ( int, const char * );
    void UnLoadDisk ( int );

    void Activate ();
    void DeActivate ();

    void WriteCRU ( ADDRESS, USHORT );
    int  ReadCRU ( ADDRESS );

    void SaveImage ( FILE * );
    void LoadImage ( FILE * );

};

#endif
