//----------------------------------------------------------------------------
//
// File:        compiler.cpp
// Date:        02-Feb-2000
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 2000-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"

#define FLAGS_XO		0x0200
#define FLAGS_OP		0x0400
#define FLAGS_OV		0x0800
#define FLAGS_CA		0x1000
#define FLAGS_EQ		0x2000
#define FLAGS_AG		0x4000
#define FLAGS_LG		0x8000

struct sCompileInfo {
    bool      isConditional;
    USHORT    nextPC;
    int       flags;
    int       clocks;
    int       pass;
    char     *source;
    char    **code;
};	  

struct sOpCode {
    USHORT    opCode;
    USHORT    mask;
    USHORT    format;
    int      (*function) ( USHORT, USHORT, sCompileInfo * );
    ULONG     clocks;
    ULONG     flags;
};	  

struct sLookUp {
    sOpCode    *opCode;
    int         size;
};

int compile_A ( USHORT, USHORT, sCompileInfo * );
int compile_AB ( USHORT, USHORT, sCompileInfo * );
int compile_ABS ( USHORT, USHORT, sCompileInfo * );
int compile_AI ( USHORT, USHORT, sCompileInfo * );
int compile_ANDI ( USHORT, USHORT, sCompileInfo * );
int compile_B ( USHORT, USHORT, sCompileInfo * );
int compile_BL ( USHORT, USHORT, sCompileInfo * );
int compile_BLWP ( USHORT, USHORT, sCompileInfo * );
int compile_C ( USHORT, USHORT, sCompileInfo * );
int compile_CB ( USHORT, USHORT, sCompileInfo * );
int compile_CI ( USHORT, USHORT, sCompileInfo * );
int compile_CKOF ( USHORT, USHORT, sCompileInfo * );
int compile_CKON ( USHORT, USHORT, sCompileInfo * );
int compile_CLR ( USHORT, USHORT, sCompileInfo * );
int compile_COC ( USHORT, USHORT, sCompileInfo * );
int compile_CZC ( USHORT, USHORT, sCompileInfo * );
int compile_DEC ( USHORT, USHORT, sCompileInfo * );
int compile_DECT ( USHORT, USHORT, sCompileInfo * );
int compile_DIV ( USHORT, USHORT, sCompileInfo * );
int compile_IDLE ( USHORT, USHORT, sCompileInfo * );
int compile_INC ( USHORT, USHORT, sCompileInfo * );
int compile_INCT ( USHORT, USHORT, sCompileInfo * );
int compile_INV ( USHORT, USHORT, sCompileInfo * );
int compile_JEQ ( USHORT, USHORT, sCompileInfo * );
int compile_JGT ( USHORT, USHORT, sCompileInfo * );
int compile_JH ( USHORT, USHORT, sCompileInfo * );
int compile_JHE ( USHORT, USHORT, sCompileInfo * );
int compile_JL ( USHORT, USHORT, sCompileInfo * );
int compile_JLE ( USHORT, USHORT, sCompileInfo * );
int compile_JLT ( USHORT, USHORT, sCompileInfo * );
int compile_JMP ( USHORT, USHORT, sCompileInfo * );
int compile_JNC ( USHORT, USHORT, sCompileInfo * );
int compile_JNE ( USHORT, USHORT, sCompileInfo * );
int compile_JNO ( USHORT, USHORT, sCompileInfo * );
int compile_JOC ( USHORT, USHORT, sCompileInfo * );
int compile_JOP ( USHORT, USHORT, sCompileInfo * );
int compile_LDCR ( USHORT, USHORT, sCompileInfo * );
int compile_LI ( USHORT, USHORT, sCompileInfo * );
int compile_LIMI ( USHORT, USHORT, sCompileInfo * );
int compile_LREX ( USHORT, USHORT, sCompileInfo * );
int compile_LWPI ( USHORT, USHORT, sCompileInfo * );
int compile_MOV ( USHORT, USHORT, sCompileInfo * );
int compile_MOVB ( USHORT, USHORT, sCompileInfo * );
int compile_MPY ( USHORT, USHORT, sCompileInfo * );
int compile_NEG ( USHORT, USHORT, sCompileInfo * );
int compile_ORI ( USHORT, USHORT, sCompileInfo * );
int compile_RSET ( USHORT, USHORT, sCompileInfo * );
int compile_RTWP ( USHORT, USHORT, sCompileInfo * );
int compile_S ( USHORT, USHORT, sCompileInfo * );
int compile_SB ( USHORT, USHORT, sCompileInfo * );
int compile_SBO ( USHORT, USHORT, sCompileInfo * );
int compile_SBZ ( USHORT, USHORT, sCompileInfo * );
int compile_SETO ( USHORT, USHORT, sCompileInfo * );
int compile_SLA ( USHORT, USHORT, sCompileInfo * );
int compile_SOC ( USHORT, USHORT, sCompileInfo * );
int compile_SOCB ( USHORT, USHORT, sCompileInfo * );
int compile_SRA ( USHORT, USHORT, sCompileInfo * );
int compile_SRC ( USHORT, USHORT, sCompileInfo * );
int compile_SRL ( USHORT, USHORT, sCompileInfo * );
int compile_STCR ( USHORT, USHORT, sCompileInfo * );
int compile_STST ( USHORT, USHORT, sCompileInfo * );
int compile_STWP ( USHORT, USHORT, sCompileInfo * );
int compile_SWPB ( USHORT, USHORT, sCompileInfo * );
int compile_SZC ( USHORT, USHORT, sCompileInfo * );
int compile_SZCB ( USHORT, USHORT, sCompileInfo * );
int compile_TB ( USHORT, USHORT, sCompileInfo * );
int compile_X ( USHORT, USHORT, sCompileInfo * );
int compile_XOP ( USHORT, USHORT, sCompileInfo * );
int compile_XOR ( USHORT, USHORT, sCompileInfo * );

extern "C" UCHAR  CpuMemory [ 0x10000 ];
extern USHORT DisassembleASM ( USHORT PC, UCHAR *, char * );

static bool used = false;
static char lineBuffer [ 256 ];
static sCompileInfo *CompiledCode [ 0x10000 >> 1 ];
static sLookUp LookUp [ 16 ];
static sOpCode OpCodes [ 69 ] = {
  { /* A   */ 0xA000, 0xF000, 1, compile_A    , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* AB  */ 0xB000, 0xF000, 1, compile_AB   , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV | FLAGS_OP },
  { /* ABS */ 0x0740, 0xFFC0, 6, compile_ABS  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OV },
  { /* AI  */ 0x0220, 0xFFE0, 8, compile_AI   , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* ANDI*/ 0x0240, 0xFFE0, 8, compile_ANDI , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ  },
  { /* B   */ 0x0440, 0xFFC0, 6, compile_B    ,  8, 0 },
  { /* BL  */ 0x0680, 0xFFC0, 6, compile_BL   , 12, 0 },
  { /* BLWP*/ 0x0400, 0xFFC0, 6, compile_BLWP , 26, 0 },
  { /* C   */ 0x8000, 0xF000, 1, compile_C    , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* CB  */ 0x9000, 0xF000, 1, compile_CB   , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* CI  */ 0x0280, 0xFFE0, 8, compile_CI   , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* CKOF*/ 0x03C0, 0xFFFF, 7, compile_CKOF , 12, 0 },
  { /* CKON*/ 0x03A0, 0xFFFF, 7, compile_CKON , 12, 0 },
  { /* CLR */ 0x04C0, 0xFFC0, 6, compile_CLR  , 10, 0 },
  { /* COC */ 0x2000, 0xFC00, 3, compile_COC  , 14, FLAGS_EQ },
  { /* CZC */ 0x2400, 0xFC00, 3, compile_CZC  , 14, FLAGS_EQ },
  { /* DEC */ 0x0600, 0xFFC0, 6, compile_DEC  , 10, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* DECT*/ 0x0640, 0xFFC0, 6, compile_DECT , 10, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* DIV */ 0x3C00, 0xFC00, 9, compile_DIV  , 16, FLAGS_OV },
  { /* IDLE*/ 0x0340, 0xFFFF, 7, compile_IDLE , 12, 0 },
  { /* INC */ 0x0580, 0xFFC0, 6, compile_INC  , 10, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* INCT*/ 0x05C0, 0xFFC0, 6, compile_INCT , 10, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* INV */ 0x0540, 0xFFC0, 6, compile_INV  , 10, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* JEQ */ 0x1300, 0xFF00, 2, compile_JEQ  , 10, 0 },
  { /* JGT */ 0x1500, 0xFF00, 2, compile_JGT  , 10, 0 },
  { /* JH  */ 0x1B00, 0xFF00, 2, compile_JH   , 10, 0 },
  { /* JHE */ 0x1400, 0xFF00, 2, compile_JHE  , 10, 0 },
  { /* JL  */ 0x1A00, 0xFF00, 2, compile_JL   , 10, 0 },
  { /* JLE */ 0x1200, 0xFF00, 2, compile_JLE  , 10, 0 },
  { /* JLT */ 0x1100, 0xFF00, 2, compile_JLT  , 10, 0 },
  { /* JMP */ 0x1000, 0xFF00, 2, compile_JMP  , 10, 0 },
  { /* JNC */ 0x1700, 0xFF00, 2, compile_JNC  , 10, 0 },
  { /* JNE */ 0x1600, 0xFF00, 2, compile_JNE  , 10, 0 },
  { /* JNO */ 0x1900, 0xFF00, 2, compile_JNO  , 10, 0 },
  { /* JOC */ 0x1800, 0xFF00, 2, compile_JOC  , 10, 0 },
  { /* JOP */ 0x1C00, 0xFF00, 2, compile_JOP  , 10, 0 },
  { /* LDCR*/ 0x3000, 0xFC00, 4, compile_LDCR , 52, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* LI  */ 0x0200, 0xFFE0, 8, compile_LI   , 12, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* LIMI*/ 0x0300, 0xFFE0, 8, compile_LIMI , 16, 0 },
  { /* LREX*/ 0x03E0, 0xFFFF, 7, compile_LREX , 12, 0 },
  { /* LWPI*/ 0x02E0, 0xFFE0, 8, compile_LWPI , 10, 0 },
  { /* MOV */ 0xC000, 0xF000, 1, compile_MOV  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* MOVB*/ 0xD000, 0xF000, 1, compile_MOVB , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* MPY */ 0x3800, 0xFC00, 9, compile_MPY  , 52, 0 },
  { /* NEG */ 0x0500, 0xFFC0, 6, compile_NEG  , 12, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OV },
  { /* ORI */ 0x0260, 0xFFE0, 8, compile_ORI  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ  },
  { /* RSET*/ 0x0360, 0xFFFF, 7, compile_RSET , 12, 0 },
  { /* RTWP*/ 0x0380, 0xFFFF, 7, compile_RTWP , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV | FLAGS_OP | FLAGS_XO },
  { /* S   */ 0x6000, 0xF000, 1, compile_S    , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* SB  */ 0x7000, 0xF000, 1, compile_SB   , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV | FLAGS_OP },
  { /* SBO */ 0x1D00, 0xFF00, 2, compile_SBO  , 12, 0 },
  { /* SBZ */ 0x1E00, 0xFF00, 2, compile_SBZ  , 12, 0 },
  { /* SETO*/ 0x0700, 0xFFC0, 6, compile_SETO , 10, 0 },
  { /* SLA */ 0x0A00, 0xFF00, 5, compile_SLA  , 52, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA | FLAGS_OV },
  { /* SOC */ 0xE000, 0xF000, 1, compile_SOC  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* SOCB*/ 0xF000, 0xF000, 1, compile_SOCB , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* SRA */ 0x0800, 0xFF00, 5, compile_SRA  , 52, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA },
  { /* SRC */ 0x0B00, 0xFF00, 5, compile_SRC  , 52, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA },
  { /* SRL */ 0x0900, 0xFF00, 5, compile_SRL  , 52, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_CA },
  { /* STCR*/ 0x3400, 0xFC00, 4, compile_STCR , 60, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* STST*/ 0x02C0, 0xFFE0, 8, compile_STST ,  8, 0 },
  { /* STWP*/ 0x02A0, 0xFFE0, 8, compile_STWP ,  8, 0 },
  { /* SWPB*/ 0x06C0, 0xFFC0, 6, compile_SWPB , 10, 0 },
  { /* SZC */ 0x4000, 0xF000, 1, compile_SZC  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ },
  { /* SZCB*/ 0x5000, 0xF000, 1, compile_SZCB , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ | FLAGS_OP },
  { /* TB  */ 0x1F00, 0xFF00, 2, compile_TB   , 12, 0 /* ???? */ },
  { /* X   */ 0x0480, 0xFFC0, 6, compile_X    ,  8, 0 },
  { /* XOP */ 0x2C00, 0xFC00, 9, compile_XOP  , 36, FLAGS_XO },
  { /* XOR */ 0x2800, 0xFC00, 3, compile_XOR  , 14, FLAGS_LG | FLAGS_AG | FLAGS_EQ } 
};

int GetArgumentB ( USHORT *PC, USHORT opCode, sCompileInfo *pInfo, int line )
{
    USHORT base;
    int reg = opCode & 0x000F;
    switch ( opCode & 0x0030 ) {
        case 0x0000 :
            sprintf ( lineBuffer, "\tlea\tecx, [ _WP_ + 2 * %d + 1 ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
        case 0x0010 :
            sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + 2 * %d ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            sprintf ( lineBuffer, "\txor\tecx, 1" );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
        case 0x0020 :
            base = * ( USHORT * ) &CpuMemory [ *PC ];
            *PC += ( USHORT ) 2;
            if ( reg == 0 ) {
                sprintf ( lineBuffer, "\tmov\tecx, 0%04Xh", base ^ 1 );
                pInfo->code [line++] = strdup ( lineBuffer );
            } else {
                sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + 2 * %d ]", reg );
                pInfo->code [line++] = strdup ( lineBuffer );
                sprintf ( lineBuffer, "\tlea\tecx, [ 0%04Xh + ecx ]", base ^ 1 );
                pInfo->code [line++] = strdup ( lineBuffer );
            }
            break;
        case 0x0030 :
            sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + 2 * %d ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            sprintf ( lineBuffer, "\tinc\t[ CpuMemoryW + _WP_ + 2 * %d ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            sprintf ( lineBuffer, "\txor\tecx, 1" );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
    }
    
    return line;
}

int GetArgumentW ( USHORT *PC, USHORT opCode, sCompileInfo *pInfo, int line )
{
    USHORT base;
    int reg = opCode & 0x000F;
    switch ( opCode & 0x0030 ) {
        case 0x0000 :
            sprintf ( lineBuffer, "\tlea\tecx, [ _WP_ + %d * 2 ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
        case 0x0010 :
            sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + %d * 2 ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
        case 0x0020 :
            base = * ( USHORT * ) &CpuMemory [ *PC ];
            *PC += ( USHORT ) 2;
            if ( reg == 0 ) {
                sprintf ( lineBuffer, "\tmov\tecx, 0%04Xh", base );
                pInfo->code [line++] = strdup ( lineBuffer );
            } else {
                sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + %d * 2 ]", reg );
                pInfo->code [line++] = strdup ( lineBuffer );
                sprintf ( lineBuffer, "\tlea\tecx, [ 0%04Xh + ecx ]", base );
                pInfo->code [line++] = strdup ( lineBuffer );
            }
            break;
        case 0x0030 :
            sprintf ( lineBuffer, "\tmov\tecx, [ CpuMemoryW + _WP_ + %d * 2 ]", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            sprintf ( lineBuffer, "\tadd\t[ CpuMemoryW + _WP_ + %d * 2 ], 2", reg );
            pInfo->code [line++] = strdup ( lineBuffer );
            break;
    }

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: A         Format: I      Op-code: 0xA000     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_A ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	mov	edi, ebx
	shr	eax, 6
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	add	bx, di
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
	mov	[ CpuMemoryW + ecx ], bx
*/
    int line = 1;

    line = GetArgumentW ( &PC, opCode, pInfo, line );

    pInfo->code [line++] = strdup ( "\tmovzx\teax, [ CpuMemoryW + ecx ]" );

    opCode >>= 6;

    line = GetArgumentW ( &PC, opCode, pInfo, line );

    pInfo->code [line++] = strdup ( "\tmovzx\tebx, [ CpuMemoryW + ecx ]" );
    pInfo->code [line++] = strdup ( "\tadd\tbx, ax" );

    pInfo->code [line++] = strdup ( "G\tmov\t[ srcComp ], ebx" );
    pInfo->code [line++] = strdup ( "G\tmov\t[ dstComp ], 0" );
    pInfo->code [line++] = strdup ( "E\tsete\t[ flagEqual ]" );
    pInfo->code [line++] = strdup ( "C\tsetc\t[ flagCarry ]" );
    pInfo->code [line++] = strdup ( "O\tseto\t[ flagOverflow ]" );

    pInfo->code [line++] = strdup ( "\tmov\t[ CpuMemoryW + ecx ], bx" );
    
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: AB        Format: I      Op-code: 0xB000     Status: L A E C O P -
//------------------------------------------------------------------------------
int compile_AB ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	shr	eax, 6
	mov	edi, ebx
	call	GetArgumentB
	call	_ReadMemoryB
	add	ebx, edi
	movsx	bx, bl
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	setpo	[ flagParity ]
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
	call	_WriteMemoryB
*/
    int line = 1;
    
    line = GetArgumentB ( &PC, opCode, pInfo, line );

    pInfo->code [line++] = strdup ( "\tcall\t_ReadMemoryB" );
    pInfo->code [line++] = strdup ( "\tmov\teax, ebx" );

    opCode >>= 6;

    line = GetArgumentB ( &PC, opCode, pInfo, line );

    pInfo->code [line++] = strdup ( "\tcall\t_ReadMemoryB" );

    pInfo->code [line++] = strdup ( "\tadd\tbx, ax" );
    pInfo->code [line++] = strdup ( "\tmovsx\tbx, bl" );

    pInfo->code [line++] = strdup ( "G\tmov\t[ srcComp ], ebx" );
    pInfo->code [line++] = strdup ( "G\tmov\t[ dstComp ], 0" );
    pInfo->code [line++] = strdup ( "E\tsete\t[ flagEqual ]" );
    pInfo->code [line++] = strdup ( "C\tsetc\t[ flagCarry ]" );
    pInfo->code [line++] = strdup ( "O\tseto\t[ flagOverflow ]" );
    pInfo->code [line++] = strdup ( "P\tseto\t[ flagParity ]" );

    pInfo->code [line++] = strdup ( "\tcall\t_WriteMemoryB" );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: ABS       Format: VI     Op-code: 0x0740     Status: L A E - O - -
//------------------------------------------------------------------------------
int compile_ABS ( USHORT, USHORT, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	cmp	bx, 0
	setl	[ flagOverflow ]
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
	jge	@@Done
	neg	[ CpuMemoryW + ecx ]
@@Done:
*/
    int line = 1;

    pInfo->code [line++] = strdup ( "G\tmov\t[ srcComp ], ebx" );
    pInfo->code [line++] = strdup ( "G\tmov\t[ dstComp ], 0" );
    pInfo->code [line++] = strdup ( "E\tsete\t[ flagEqual ]" );
    pInfo->code [line++] = strdup ( "O\tseto\t[ flagOverflow ]" );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: AI        Format: VIII   Op-code: 0x0220     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_AI ( USHORT, USHORT, sCompileInfo *pInfo )
{
/*
	mov	ecx, eax
	and	ecx, 0000000Fh
	mov	bx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	mov	dx, [ CpuMemoryW + _WP_ + ecx * 2 ]
	add	bx, dx
	mov	[ CpuMemoryW + _WP_ + ecx * 2 ], bx
	setc	[ flagCarry ];
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;

    pInfo->code [line++] = strdup ( "G\tmov\t[ srcComp ], ebx" );
    pInfo->code [line++] = strdup ( "G\tmov\t[ dstComp ], 0" );
    pInfo->code [line++] = strdup ( "E\tsete\t[ flagEqual ]" );
    pInfo->code [line++] = strdup ( "C\tsetc\t[ flagCarry ]" );
    pInfo->code [line++] = strdup ( "O\tseto\t[ flagOverflow ]" );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: ANDI      Format: VIII   Op-code: 0x0240     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_ANDI ( USHORT, USHORT, sCompileInfo *pInfo )
{
/*
	and	eax, 0000000Fh
	movzx	ebx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	mov	dx, [ CpuMemoryW + _WP_ + eax * 2 ]
	and	bx, dx
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;

    pInfo->code [line++] = strdup ( "G\tmov\t[ srcComp ], ebx" );
    pInfo->code [line++] = strdup ( "G\tmov\t[ dstComp ], 0" );
    pInfo->code [line++] = strdup ( "E\tsete\t[ flagEqual ]" );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: B         Format: VI     Op-code: 0x0440     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_B ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
	mov	_PC_, ecx
*/
    int line = 1;

    if (( opCode & 0x0030 ) == 0x0020 ) {
        pInfo->nextPC = * ( USHORT * ) &CpuMemory [ PC ];
        sprintf ( lineBuffer, "\tjmp\tL%04X", pInfo->nextPC );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        pInfo->isConditional = true;
        line = GetArgumentW ( &PC, opCode, pInfo, line );
        pInfo->code [line++] = strdup ( "\tjmp\t[ JumpTable + ecx * 2 ]" );
    }
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: BL        Format: VI     Op-code: 0x0680     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_BL ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
	mov	[ CpuMemoryW + _WP_ + 11 * 2 ], _PCw_
	mov	_PC_, ecx
*/
    int line = 1;

    if (( opCode & 0x0030 ) == 0x0020 ) {
        pInfo->nextPC = * ( USHORT * ) &CpuMemory [ PC ];
        sprintf ( lineBuffer, "\tmov\t[ CpuMemoryW + _WP_ + 11 * 2 ], 0%04XH", PC + 2 );
        pInfo->code [line++] = strdup ( lineBuffer );
        sprintf ( lineBuffer, "\tjmp\tL%04X", pInfo->nextPC );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        pInfo->isConditional = true;
        sprintf ( lineBuffer, "\tmov\t[ CpuMemoryW + _WP_ + 11 * 2 ], 0%04XH", PC );
        pInfo->code [line++] = strdup ( lineBuffer );
        // Jump to L%ecx%!!
    }


    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: BLWP      Format: VI     Op-code: 0x0400     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_BLWP ( USHORT PC, USHORT, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
	movzx	eax, [ CpuMemoryW + ecx ]
	movzx	ebx, [ CpuMemoryW + ecx + 2 ]
	call	ContextSwitch
*/
    int line = 1;

    pInfo->nextPC = CpuMemory [ PC ];

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: C         Format: I      Op-code: 0x8000     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_C ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edi, [ CpuMemoryW + ecx ]
	shr	eax, 6
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	mov	[ srcComp ], edi
	mov	[ dstComp ], ebx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CB        Format: I      Op-code: 0x9000     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_CB ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	cmp     bl, 0
	setpo	[ flagParity ]
	mov	[ srcComp ], ebx
	shr	eax, 6
	call	GetArgumentB
	call	_ReadMemoryB
	mov	[ dstComp ], ebx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CI        Format: VIII   Op-code: 0x0280     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_CI ( USHORT, USHORT, sCompileInfo * )
{
/*
	and	eax, 0000000Fh
	movzx	edx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	movzx	ebx, [ CpuMemoryW + _WP_ + eax * 2 ]
	mov	[ srcComp ], ebx
	mov	[ dstComp ], edx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CKOF      Format: VII    Op-code: 0x03C0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_CKOF ( USHORT, USHORT, sCompileInfo * )
{
/*
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CKON      Format: VII    Op-code: 0x03A0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_CKON ( USHORT, USHORT, sCompileInfo * )
{
/*
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CLR       Format: VI     Op-code: 0x04C0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_CLR ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	xor	ebx, ebx
	call	_WriteMemoryW
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: COC       Format: III    Op-code: 0x2000     Status: - - E - - - -
//------------------------------------------------------------------------------
int compile_COC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	shr	eax, 6
	and	eax, 0000000Fh
	mov	edi, ebx
	and	bx, [ CpuMemoryW + _WP_ + eax * 2 ]
	mov	[ srcComp ], ebx
	mov	[ dstComp ], edi
*/
    // Shouldn't affect LG & AG!

    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: CZC       Format: III    Op-code: 0x2400     Status: - - E - - - -
//------------------------------------------------------------------------------
int compile_CZC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	shr	eax, 6
	and	eax, 0000000Fh
	mov	di, [ CpuMemoryW + _WP_ + eax * 2 ]
	not	di
	and	di, bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], edi
*/
    // Shouldn't affect LG & AG!

    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: DEC       Format: VI     Op-code: 0x0600     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_DEC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	add	bx, -1
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: DECT      Format: VI     Op-code: 0x0640     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_DECT ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	add	bx, -2
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: DIV       Format: IX     Op-code: 0x3C00     Status: - - - - O - -
//------------------------------------------------------------------------------
int compile_DIV ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edi, [ CpuMemoryW + ecx ]
	shr	eax, 6
	and	al, 0Fh
	call	GetArgumentW
	movzx	edx, [ CpuMemoryW + ecx ]
	cmp	edx, edi
	jg	@@overflow
	movzx	eax, [ CpuMemoryW + ecx + 2 ]
	div	di
	mov	[ CpuMemoryW + ecx ], ax		; quotient
	mov	[ CpuMemoryW + ecx + 2 ], dx		; remainder
	mov	[ flagOverflow ], 0
	ret

@@overflow:
	mov	[ flagOverflow ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: IDLE      Format: VII    Op-code: 0x0340     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_IDLE ( USHORT, USHORT, sCompileInfo * )
{
/*
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: INC       Format: VI     Op-code: 0x0580     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_INC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	add	bx, 1
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: INCT      Format: VI     Op-code: 0x05C0     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_INCT ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	add	bx, 2
	setc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: INV       Format: VI     Op-code: 0x0540     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_INV ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	not	ebx
	call	_WriteMemoryW
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JEQ       Format: II     Op-code: 0x1300     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JEQ ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tje\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JGT       Format: II     Op-code: 0x1500     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JGT ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjg\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JH        Format: II     Op-code: 0x1B00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JH ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tja\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );
		 
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JHE       Format: II     Op-code: 0x1400     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JHE ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjae\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}		 

//------------------------------------------------------------------------------
// Mnemonic: JL        Format: II     Op-code: 0x1A00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JL ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjb\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}		 

//------------------------------------------------------------------------------
// Mnemonic: JLE       Format: II     Op-code: 0x1200     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JLE ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjbe\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}		 

//------------------------------------------------------------------------------
// Mnemonic: JLT       Format: II     Op-code: 0x1100     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JLT ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjl\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}		 

//------------------------------------------------------------------------------
// Mnemonic: JMP       Format: II     Op-code: 0x1000     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JMP ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjmp\tL%04X", PC );

    pInfo->nextPC = PC;
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JNC       Format: II     Op-code: 0x1700     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JNC ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tje\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tcmp\t[ flagCarry ], 0" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JNE       Format: II     Op-code: 0x1600     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JNE ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjne\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tmov\tebx, [ srcComp ]" );
    pInfo->code [line++] = strdup ( "\tcmp\tbx, [ word ptr dstComp ]" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}		 

//------------------------------------------------------------------------------
// Mnemonic: JNO       Format: II     Op-code: 0x1900     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JNO ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tje\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tcmp\t[ flagOverflow ], 0" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JOC       Format: II     Op-code: 0x1800     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JOC ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjne\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tcmp\t[ flagCarry ], 0" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: JOP       Format: II     Op-code: 0x1C00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_JOP ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
*/
    int line = 1;

    PC = ( USHORT ) ( PC + ( opCode & 0xFF ) * 2 );
    sprintf ( lineBuffer, "\tjne\tL%04X", PC );

    pInfo->isConditional = true;
    pInfo->code [line++] = strdup ( "\tcmp\t[ flagParity ], 0" );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: LDCR      Format: IV     Op-code: 0x3000     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_LDCR ( USHORT, USHORT, sCompileInfo * )
{
/*
;    USHORT address = GetSrcAddress () & 0xFFFE;
;    USHORT value = ReadMemoryW ( address );

;    int cru = ReadRegisterW ( 12 ) >> 1;

	movzx	edi, [ CpuMemoryW + _WP_ + 12 * 2 ]
	shr	edi, 1
	and	edi, 00000FFFh

;    int count = ( curOpCode >> 6 ) & 0x000F;

	mov	edx, eax
	shr	edx, 6
	and	edx, 0000000Fh

;    if ( count == 0 ) count = 16;
;    if ( count < 9 ) {
;        value >>= 8;
;        _Compare_ (( UCHAR ) value, 0 );
;    } else
;        _Compare_ ( value, 0 );

	cmp	edx, 0
	je	@@SetSixteen
	cmp	edx, 9
	jge	@@Word

	push	edx
	call	GetArgumentB
	movsx	ebx, [ CpuMemoryB + ecx ]

	cmp	bl, 0
	setpo	[ flagParity ]
	mov	[ srcComp ], ebx
	jmp	@@Write
        
@@SetSixteen:
	mov	edx, 16
@@Word:
	push	edx
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	mov	[ srcComp ], ebx

;    while ( count-- ) {
;        WriteCRU ( cru++, ( value & 1 ) ? 1 : 0 );
;        value >>= 1;
;    }

@@Write:
	pop	edx
	mov	[ dstComp ], 0

	push	esi

	push	ebx
	push	edx
	push	edi
	push	[_CRU_Object]
	call	_WriteCRU
	pop	ebx
	pop	ebx
	pop	ebx
	pop	ebx

	pop	esi
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: LI        Format: VIII   Op-code: 0x0200     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_LI ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	and	eax, 0000000Fh
	movzx	ebx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;

    int reg = opCode & 0x000F;
    USHORT imm = * ( USHORT * ) &CpuMemory [ PC ];
    sprintf ( lineBuffer, "\tmov\t[ CpuMemoryW + _WP_ + %d * 2 ], 0%04Xh", reg, imm );
    pInfo->code [line++] = strdup ( lineBuffer );

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: LIMI      Format: VIII   Op-code: 0x200      Status: L A E - - - -
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Mnemonic: LIMI      Format: VIII   Op-code: 0x0300     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_LIMI ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	al, [ CpuMemoryB + _PC_ ]
	mov	[ intLevel ], al
	add	_PCw_, 2
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: LREX      Format: VII    Op-code: 0x03E0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_LREX ( USHORT, USHORT, sCompileInfo * )
{
/*
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: LWPI      Format: VIII   Op-code: 0x02E0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_LWPI ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	mov	_WP_, ecx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: MOV       Format: I      Op-code: 0xC000     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_MOV ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentW
        call    _ReadMemoryW
	shr	eax, 6
	call	GetArgumentW
        call    _WriteMemoryW
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    
    if ( opCode & 0x0030 ) {
        sprintf ( lineBuffer, "\tmovzx\teax, [ CpuMemoryW + _WP_ + %d * 2 ]", opCode & 0x0F );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        line = GetArgumentW ( &PC, opCode, pInfo, line );
        pInfo->code [line++] = strdup ( "\tcall\t_ReadMemoryW" );
    }

    opCode >>= 6;

    if ( opCode & 0x0030 ) {
        sprintf ( lineBuffer, "\tmov\t[ CpuMemoryW + _WP_ + %d * 2 ], ax", opCode & 0x0F );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        line = GetArgumentW ( &PC, opCode, pInfo, line );
        pInfo->code [line++] = strdup ( "\tcall\t_WriteMemoryW" );
    }

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: MOVB      Format: I      Op-code: 0xD000     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_MOVB ( USHORT PC, USHORT opCode, sCompileInfo *pInfo )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	shr	eax, 6
	call	GetArgumentB
	call	_WriteMemoryB
	cmp	bl, 0
	setpo	[ flagParity ]
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    
    if ( opCode & 0x0030 ) {
        sprintf ( lineBuffer, "\tmovsx\teax, [ CpuMemoryB + _WP_ + %d * 2 + 1 ]", opCode & 0x0F );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        line = GetArgumentB ( &PC, opCode, pInfo, line );
        pInfo->code [line++] = strdup ( "\tcall\t_ReadMemoryB" );
    }

    opCode >>= 6;

    if ( opCode & 0x0030 ) {
        sprintf ( lineBuffer, "\tmov\t[ CpuMemoryB + _WP_ + %d * 2 + 1 ], al", opCode & 0x0F );
        pInfo->code [line++] = strdup ( lineBuffer );
    } else {
        line = GetArgumentB ( &PC, opCode, pInfo, line );
        pInfo->code [line++] = strdup ( "\tcall\t_WriteMemoryB" );
    }

    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: MPY       Format: IX     Op-code: 0x3800     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_MPY ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edx, [ CpuMemoryW + ecx ]
	shr	eax, 6
	and	al, 0Fh
	call	GetArgumentW
	movzx	eax, [ CpuMemoryW + ecx ]
	mul	dx
	mov	[ CpuMemoryW + ecx ], dx
	mov	[ CpuMemoryW + ecx + 2 ], ax
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: NEG       Format: VI     Op-code: 0x0500     Status: L A E - O - -
//------------------------------------------------------------------------------
int compile_NEG ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	neg	bx
	seto	[ flagOverflow ];
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: ORI       Format: VIII   Op-code: 0x0260     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_ORI ( USHORT, USHORT, sCompileInfo * )
{
/*
	and	eax, 0000000Fh
	movzx	ebx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	mov	dx, [ CpuMemoryW + _WP_ + eax * 2 ]
	or	bx, dx
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: RSET      Format: VII    Op-code: 0x0360     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_RSET ( USHORT, USHORT, sCompileInfo * )
{
/*
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: RTWP      Format: VII    Op-code: 0x0380     Status: L A E C O P X
//------------------------------------------------------------------------------
int compile_RTWP ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	dx, [ CpuMemoryW + _WP_ + 15 * 2 ]
	mov	_PCw_, [ CpuMemoryW + _WP_ + 14 * 2 ]
	mov	_WPw_, [ CpuMemoryW + _WP_ + 13 * 2 ]
	call	SetStatus
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: S         Format: I      Op-code: 0x6000     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_S ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edi, [ CpuMemoryW + ecx ]
	shr	eax, 6
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	sub	bx, di
	setnc	[ flagCarry ]
	seto	[ flagOverflow ]
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SB        Format: I      Op-code: 0x7000     Status: L A E C O P -
//------------------------------------------------------------------------------
int compile_SB ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	shr	eax, 6
	mov	ah, bl
	call	GetArgumentB
	call	_ReadMemoryB
	sub	bl, ah
	movsx	bx, bl
	setnc	[ flagCarry ]
	seto	[ flagOverflow ]
	setpo	[ flagParity ]
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
	call	_WriteMemoryB
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SBO       Format: II     Op-code: 0x1D00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_SBO ( USHORT, USHORT, sCompileInfo * )
{
/*
	movzx	ebx, [ CpuMemoryW + _WP_ + 12 * 2 ]
	shr	ebx, 1
	and	eax, 000000FFh
	add	ebx, eax
	push	esi
	push	edx
	push	1
	push	1
	push	ebx
	push	[_CRU_Object]
	call	_WriteCRU
	pop	ebx
	pop	ebx
	pop	ebx
	pop	ebx
	pop	edx
	pop	esi
	and	al, 1
	shl	al, 5
	or	dl, al
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SBZ       Format: II     Op-code: 0x1E00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_SBZ ( USHORT, USHORT, sCompileInfo * )
{
/*
	movzx	ebx, [ CpuMemoryW + _WP_ + 12 * 2 ]
	shr	ebx, 1
	and	eax, 000000FFh
	add	ebx, eax
	push	esi
	push	edx
	push	0
	push	1
	push	ebx
	push	[_CRU_Object]
	call	_WriteCRU
	pop	ebx
	pop	ebx
	pop	ebx
	pop	ebx
	pop	edx
	pop	esi
	and	al, 1
	shl	al, 5
	or	dl, al
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SETO      Format: VI     Op-code: 0x0700     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_SETO ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	mov	ebx, -1
	mov	[ CpuMemoryW + ecx ], bx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SLA       Format: V      Op-code: 0x0A00     Status: L A E C O - -
//------------------------------------------------------------------------------
int compile_SLA ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	ecx, eax
	and	eax, 0000000Fh
	shr	ecx, 4
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	cx, [ CpuMemoryW + _WP_ + 0 * 2 ]
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	ecx, 16

@@ValidShift:

	movsx	ebx, [ CpuMemoryW + _WP_ + eax * 2 ]
	sal	ebx, cl
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
	shr	ebx, 16
	je	@@NoBits
	test	ebx, 1
	setne	[ flagCarry ]
	inc	ebx
	setne	[ flagOverflow ]
	ret

@@NoBits:

	mov	[ flagCarry ], 0
	mov	[ flagOverflow ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SOC       Format: I      Op-code: 0xE000     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_SOC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edi, [ CpuMemoryW + ecx ]
	shr	eax, 6
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	or	bx, di
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
	mov	[ CpuMemoryW + ecx ], bx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SOCB      Format: I      Op-code: 0xF000     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_SOCB ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	shr	eax, 6
	mov	ah, bl
	call	GetArgumentB
	call	_ReadMemoryB
	or	bl, ah
	movsx	bx, bl
	setpo	[ flagParity ]
	mov 	[ srcComp ], ebx
	mov	[ dstComp ], 0
	call	_WriteMemoryB
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SRA       Format: V      Op-code: 0x0800     Status: L A E C - - -
//------------------------------------------------------------------------------
int compile_SRA ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	ecx, eax
	and	eax, 0000000Fh
	shr	ecx, 4
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	cx, [ CpuMemoryW + _WP_ + 0 * 2 ]
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	ecx, 16

@@ValidShift:

	movsx	ebx, [ CpuMemoryW + _WP_ + eax * 2 ]
	sar	ebx, cl
	setc	[ flagCarry ]
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SRC       Format: V      Op-code: 0x0B00     Status: L A E C - - -
//------------------------------------------------------------------------------
int compile_SRC ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	ecx, eax
	and	eax, 0000000Fh
	shr	ecx, 4
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	cx, [ CpuMemoryW + _WP_ + 0 * 2 ]
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	ecx, 16

@@ValidShift:

	movzx	ebx, [ CpuMemoryW + _WP_ + eax * 2 ]
	ror	bx, cl
	setc	[ flagCarry ]
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SRL       Format: V      Op-code: 0x0900     Status: L A E C - - -
//------------------------------------------------------------------------------
int compile_SRL ( USHORT, USHORT, sCompileInfo * )
{
/*
	mov	ecx, eax
	and	eax, 0000000Fh
	shr	ecx, 4
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	cx, [ CpuMemoryW + _WP_ + 0 * 2 ]
	and	ecx, 0000000Fh
	jne	@@ValidShift
	mov	ecx, 16

@@ValidShift:

	movzx	ebx, [ CpuMemoryW + _WP_ + eax * 2 ]
	shr	ebx, cl
	setc	[ flagCarry ]
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: STCR      Format: IV     Op-code: 0x3400     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_STCR ( USHORT, USHORT, sCompileInfo * )
{
/*
;    USHORT address = GetSrcAddress () & 0xFFFE;

	mov	ecx, eax

;    int cru = ReadRegisterW ( 12 ) >> 1;

	movzx	edi, [ CpuMemoryW + _WP_ + 12 * 2 ]
	shr	edi, 1
	and	edi, 00000FFFh

;    int count = ( curOpCode >> 6 ) & 0x000F;

	shr	eax, 6
	and	eax, 0000000Fh

;    if ( count == 0 ) count = 16;

	sete	bl
	shl	bl, 4
	or	al, bl

;        while ( count-- ) {
;            value <<= 1;
;            value |= ReadCRU ( cru++ );
;        }

	push	esi
	push	ecx

	push	eax
	push	edi
	push	[_CRU_Object]
	call	_ReadCRU
	pop	ebx
	pop	ebx
	pop	edi

	pop	ecx
	pop	esi

;    if ( count < 9 ) {
;        _Compare_ ( value, 0 );
;        WriteMemory ( address, value );
;    } else {
;        _Compare_ ( value, 0 );
;        WriteMemory ( address, value );
;    }

        mov	edx, eax
        mov	eax, ecx
        
 	cmp	edi, 9
	jge	@@Word

	call	GetArgumentB
	push	ecx
	movzx	ebx, [ CpuMemoryB + ecx ]
	mov	ecx, edi
	shl	bl, cl
	or	bl, dl
	setpo	[ flagParity ]
	movsx	edx, bl
	mov	[ srcComp ], edx
	mov	[ dstComp ], 0
	pop	ecx
	mov	[ CpuMemoryB + ecx ], bl
	ret
@@Word:

	call	GetArgumentW
	push	ecx
	movzx	ebx, [ CpuMemoryW + ecx ]
	mov	ecx, edi
	shl	ebx, cl
	or	ebx, edx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
	pop	ecx
	mov	[ CpuMemoryW + ecx ], bx
	ret
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: STST      Format: VIII   Op-code: 0x02C0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_STST ( USHORT, USHORT, sCompileInfo * )
{
/*
	and	eax, 0000000Fh
	call	GetStatus
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], dx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: STWP      Format: VIII   Op-code: 0x02A0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_STWP ( USHORT, USHORT, sCompileInfo * )
{
/*
	and	eax, 0000000Fh
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], _WPw_
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SWPB      Format: VI     Op-code: 0x06C0     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_SWPB ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	xchg	bh, bl
	mov	[ CpuMemoryW + ecx ], bx
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SZC       Format: I      Op-code: 0x4000     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_SZC ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	edi, [ CpuMemoryW + ecx ]
	not	edi
	shr	eax, 6
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	and	bx, di
	mov	[ CpuMemoryW + ecx ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: SZCB      Format: I      Op-code: 0x5000     Status: L A E - - P -
//------------------------------------------------------------------------------
int compile_SZCB ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentB
	call	_ReadMemoryB
	not	ebx
	shr	eax, 6
	mov	ah, bl
	call	GetArgumentB
	call	_ReadMemoryB
	and	bl, ah
        movsx	bx, bl
	setpo	[ flagParity ]
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
	call	_WriteMemoryB
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: TB        Format: II     Op-code: 0x1F00     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_TB ( USHORT, USHORT, sCompileInfo * )
{
/*
	movzx	ebx, [ CpuMemoryW + _WP_ + 12 * 2 ]
	shr	ebx, 1
	and	eax, 000000FFh
	add	ebx, eax
	push	esi
	push	edx
	push	1
	push	ebx
	push	[_CRU_Object]
	call	_ReadCRU
	pop	ebx
	pop	ebx
	pop	ebx
	pop	edx
	pop	esi
	and	eax, 1
	mov	[ srcComp ], eax
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: X         Format: VI     Op-code: 0x0480     Status: - - - - - - -
//------------------------------------------------------------------------------
int compile_X ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	eax, [ CpuMemoryW + ecx ]
	call	ExecuteInstruction
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: XOP       Format: IX     Op-code: 0x2C00     Status: - - - - - - X
//------------------------------------------------------------------------------
int compile_XOP ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	mov	edi, ebx		; return Address
	shr	eax, 4
	and	eax, 0000003Ch
	lea	ecx, [ eax + 64 ]
	movzx	eax, [ CpuMemoryW + ecx ]
	movzx	ebx, [ CpuMemoryW + ecx + 2]
	call	ContextSwitch
	mov	[ CpuMemoryW + _WP_ + 11 * 2 ], di
	mov	[ flagXOP ], 1
*/
    int line = 1;
    return line;
}

//------------------------------------------------------------------------------
// Mnemonic: XOR       Format: III    Op-code: 0x2800     Status: L A E - - - -
//------------------------------------------------------------------------------
int compile_XOR ( USHORT, USHORT, sCompileInfo * )
{
/*
	call	GetArgumentW
	movzx	ebx, [ CpuMemoryW + ecx ]
	shr	eax, 6
	and	eax, 0000000Fh
	mov	dx, [ CpuMemoryW + _WP_ + eax * 2 ]
	xor	bx, dx
	mov	[ CpuMemoryW + _WP_ + eax * 2 ], bx
	mov	[ srcComp ], ebx
	mov	[ dstComp ], 0
*/
    int line = 1;
    return line;
}

sCompileInfo *CompileOpcode ( USHORT PC )
{
    sCompileInfo *pInfo = new sCompileInfo;

    memset ( pInfo, 0, sizeof ( *pInfo ));

    char buffer [ 256 ];
    int lines = 20;

    pInfo->code = new char * [ lines + 2 ];
    memset ( pInfo->code, 0, sizeof ( char ** ) * ( lines + 2 ));

    // First line is the address label
    sprintf ( buffer, "L%04X:", PC );
    pInfo->code [0] = strdup ( buffer );

    USHORT curOpCode = * ( USHORT * ) &CpuMemory [ PC ];
    sLookUp *lookup = &LookUp [ curOpCode >> 12 ];
    sOpCode *op = lookup->opCode;
    int retries = lookup->size;
    if ( retries ) {
        while (( curOpCode & op->mask ) != op->opCode ) {
            if ( retries-- == 0 ) {
                break;
            }
            op++;
        }
    }

    pInfo->nextPC = DisassembleASM ( PC, &CpuMemory [ PC ], buffer );
    pInfo->source = strdup ( buffer );
    pInfo->flags  = op->flags;
    pInfo->clocks = op->clocks;
    pInfo->pass   = 1;

    op->function ( PC + 2, curOpCode, pInfo );

    return pInfo;
}

void Compile ( USHORT PC )
{
    used = true;
    if ( CompiledCode [ PC >> 1 ] != NULL ) {
        CompiledCode [ PC >> 1 ]->pass++;
        return;
    }
    CompiledCode [ PC >> 1 ] = CompileOpcode ( PC );
}

typedef int (*QSORT_FUNC) ( const void *, const void * );

int SortFunction ( const sOpCode *p1, const sOpCode *p2 )
{
    if ( p1->opCode < p2->opCode ) return -1;
    if ( p1->opCode > p2->opCode ) return 1;
    return 0;
}

static void dump ()
{
    if ( used == false ) return;

    FILE *file = fopen ( "compiled.ti", "wt" );

    for ( int PC = 0; PC < 0x10000; PC += 2 ) {

        sCompileInfo *pInfo = CompiledCode [ PC >> 1 ];

        if ( pInfo == NULL ) continue;

        sCompileInfo *pNext = ( pInfo->isConditional == true ) ? ( sCompileInfo * ) NULL : CompiledCode [ pInfo->nextPC >> 1 ];

        bool flagLG = ( pInfo->flags & FLAGS_LG ) ? ( pNext && ( pNext->flags & FLAGS_LG )) ? false : true : false;
        bool flagAG = ( pInfo->flags & FLAGS_AG ) ? ( pNext && ( pNext->flags & FLAGS_AG )) ? false : true : false;
        bool flagEQ = ( pInfo->flags & FLAGS_EQ ) ? ( pNext && ( pNext->flags & FLAGS_EQ )) ? false : true : false;
        bool flagCA = ( pInfo->flags & FLAGS_CA ) ? ( pNext && ( pNext->flags & FLAGS_CA )) ? false : true : false;
        bool flagOV = ( pInfo->flags & FLAGS_OV ) ? ( pNext && ( pNext->flags & FLAGS_OV )) ? false : true : false;
        bool flagOP = ( pInfo->flags & FLAGS_OP ) ? ( pNext && ( pNext->flags & FLAGS_OP )) ? false : true : false;
        bool flagXO = ( pInfo->flags & FLAGS_XO ) ? ( pNext && ( pNext->flags & FLAGS_XO )) ? false : true : false;

        fprintf ( file, "; %-35.35s  Hits: %8d Next: %04X  Flags: %c%c%c%c%c%c%c\n", pInfo->source, pInfo->pass,
            ( pInfo->isConditional == true ) ? ( USHORT ) -1 : pInfo->nextPC,
            ( pInfo->flags & FLAGS_LG ) ? ( pNext && ( pNext->flags & FLAGS_LG )) ? 'l' : 'L' : '-',
            ( pInfo->flags & FLAGS_AG ) ? ( pNext && ( pNext->flags & FLAGS_AG )) ? 'a' : 'A' : '-',
            ( pInfo->flags & FLAGS_EQ ) ? ( pNext && ( pNext->flags & FLAGS_EQ )) ? 'e' : 'E' : '-',
            ( pInfo->flags & FLAGS_CA ) ? ( pNext && ( pNext->flags & FLAGS_CA )) ? 'c' : 'C' : '-',
            ( pInfo->flags & FLAGS_OV ) ? ( pNext && ( pNext->flags & FLAGS_OV )) ? 'o' : 'O' : '-',
            ( pInfo->flags & FLAGS_OP ) ? ( pNext && ( pNext->flags & FLAGS_OP )) ? 'p' : 'P' : '-',
            ( pInfo->flags & FLAGS_XO ) ? ( pNext && ( pNext->flags & FLAGS_XO )) ? 'x' : 'X' : '-' );

        char **line = pInfo->code;
        while (( line != NULL ) && ( *line != NULL )) {
            int ch = *line [0];
            bool display = false;
            switch ( ch ) {
                case 'G' :
                    display = flagLG | flagAG;
                    break;
                case 'E' :
                    display = flagEQ;
                    break;
                case 'C' :
                    display = flagCA;
                    break;
                case 'O' :
                    display = flagOV;
                    break;
                case 'P' :
                    display = flagOP;
                    break;
                case 'X' :
                    display = flagXO;
                    break;
                default :
                    fprintf ( file, "%s\n", *line );
                    break;
            }
            if ( display == true ) {
                fprintf ( file, "%s\n", *line + 1 );
            }
            line++;
        }
    }

    fclose ( file );
}

int startup ()
{
    // Sort the OpCode table by OpCode
    qsort ( OpCodes, SIZE ( OpCodes ), sizeof ( OpCodes[0] ), ( QSORT_FUNC ) SortFunction );

    // Create the LookUp table using the high 4 bits of each OpCode
    int last = 0;
    unsigned x;
    for ( x = 0; x < SIZE ( LookUp ) - 1; x++ ) {
        int index = last + 1;
        while ( x + 1 > ( unsigned ) ( OpCodes[index].opCode >> 12 )) index++;
        LookUp [x].opCode = &OpCodes [last];
        LookUp [x].size = index - last - 1;
        last = index;
    }
    LookUp [x].opCode = &OpCodes [last];
    LookUp [x].size = SIZE ( OpCodes ) - last - 1;

    atexit ( dump );

    return 0;
}

static int init = startup ();
